/*
* JBoss, Home of Professional Open Source
* Copyright 2009, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.metadata.spi.scope;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;

import org.jboss.util.collection.ConcurrentSkipListMap;

/**
 * The ScopeKey represents a path which is made up of the path entries.  Elements may be added dynamically
 * until the ScopeKey is frozen.  If you want to add more elements to a frozen ScopeKey, clone it
 * (cloned objects are not frozen).
 *
 * To use the ScopeKey as a key object, use the UnmodifiableScopeKey returned from getOptimizedKey().
 * UnmodifiableScopeKey's are immutable and suitable for use as a key (also thread-safe with no object lock contention). 
 *
 * The ScopeKey (Server=Bob,Deployment=Foo.war,Class=Bar)
 * is the child of (Server=Bob,Deployment=Foo.war),
 * which is the child of (Server=Bob).  Think about this statement in terms of
 * a path relationship.  The server (Bob) contains a deployment (Foo.war) and
 * the deployment contains a class (Bar).
 *
 * See historical design notes
 * http://www.jboss.org/index.html?module=bb&op=viewtopic&p=3972233#3972233
 * http://www.jboss.org/index.html?module=bb&op=viewtopic&p=4013747#4013747
 *
 * Thread Safety:  This class is thread safe.
 *
 * Warning:  This class is (delicately) extended by UnmodifiableScopeKey.
 * Since there is no common interface class between the two classes, there are risks to be
 * aware of.  If a new method is added to ScopeKey and not to UnmodifiableScopeKey, undesireable
 * behavior will occur when that UnmodifiableScopeKey.method is invoked (ScopeKey's member variables will be null). 
 *
 * @author <a href="adrian@jboss.com">Adrian Brock</a>
 * @version $Revision: 101076 $
 */
public class ScopeKey implements Serializable, Cloneable
{
   private final static ConcurrentSkipListMap<ScopeLevel, Scope> EMPTY_SCOPES = new UnmodifiableConcurrentSkipListMap<ScopeLevel, Scope>(); 
   
   /** The serialVersionUID */
   private static final long serialVersionUID = -496238095349593371L + 1L;

   /** The default scope */
   public static final ScopeKey DEFAULT_SCOPE = new ScopeKey(new Scope(CommonLevels.JVM, "THIS"));
   
   /** The scopes () */
   private volatile ConcurrentSkipListMap<ScopeLevel, Scope> scopes = EMPTY_SCOPES;
   
   /** The scope level for this key */
   private ScopeLevel maxScopeLevel;
   
   /** Whether the key is frozen */
   private volatile boolean frozen;

   static
   {
      DEFAULT_SCOPE.freeze();
   }
   
   /**
    * Create a new ScopeKey.
    */
   public ScopeKey()
   {
   }

   /**
    * Create a new ScopeKey.
    * 
    * @param scope the scope
    */
   public ScopeKey(Scope scope)
   {
      addScope(scope);
   }

   /**
    * Create a new ScopeKey.
    * 
    * @param level the scope level
    * @param qualifier the scope qualifier
    */
   public ScopeKey(ScopeLevel level, Object qualifier)
   {
      addScope(level, qualifier);
   }

   /**
    * Create a new ScopeKey.
    * 
    * @param scopes is a collection of Scope instances that make up the represented path
    * @throws IllegalArgumentException if parameter scopes is null
    */
   public ScopeKey(Collection<Scope> scopes)
   {
      if (scopes == null)
         throw new IllegalArgumentException("Null scopes");
      for (Scope scope : scopes)
         addScope(scope);
   }

   /**
    * Create a new ScopeKey.
    * 
    * @param scopes is zero or more Scope instances that make up the represented path
    * @throws IllegalArgumentException if parameter scopes is null
    */
   public ScopeKey(Scope[] scopes)
   {
      if (scopes == null)
         throw new IllegalArgumentException("Null scopes");
      for (Scope scope : scopes)
         addScope(scope);
   }
   
   /**
    * Get the frozen.
    * 
    * @return the frozen.
    */
   public boolean isFrozen()
   {
      return frozen;
   }

   /**
    * Set to frozen.
    *
    * @throws IllegalArgumentException if there are no Scope entries added.
    */
   public void freeze()
   {
      if (scopes.isEmpty())
         throw new IllegalStateException("Attempt to freeze an empty key");
      this.frozen = true;
   }
   
   /**
    * The returned ScopeKey is immutable and optimized for use at runtime.
    *
    * @return Optimized immutable ScopeKey
    */
   public ScopeKey getOptimizedKey()
   {
      return new UnmodifiableScopeKey(this);
   }

   /**
    * Get the scopes 
    * 
    * @return the scopes in expected path order
    */
   public Collection<Scope> getScopes()
   {
      return Collections.unmodifiableCollection(scopes.values());
   }

   /**
    * Get a scope
    * 
    * @param level the scope level
    * @return the scope
    * @throws IllegalArgumentException if level is null.
    */
   public Scope getScope(ScopeLevel level)
   {
      if (level == null)
         throw new IllegalArgumentException("Null level");
      return scopes.get(level);
   }
   
   /**
    * Get the maximum scope level
    * 
    * @return the largest scope level
    */
   public ScopeLevel getMaxScopeLevel()
   {
      return maxScopeLevel;
   }
   
   /**
    * Get the parent scope key
    * 
    * @return the parent or null if there is no parent (meaning that we
    * are at the top most element in the path)
    */
   public ScopeKey getParent()
   {
      if (scopes.size() < 2)
         return null;
      
      ScopeKey result = new ScopeKey();
      for (Iterator<Scope> i = scopes.values().iterator(); i.hasNext();)
      {
         Scope scope = i.next();
         if (i.hasNext())
            result.addScope(scope);
      }
      return result;
   }

   /**
    * Is this parent of key parameter.
    *
    * @param key the key parameter
    * @return true if this is direct parent of key param
    * @throws IllegalArgumentException if parameter key is null
    */
   public boolean isParent(ScopeKey key)
   {
      if (key == null)
         throw new IllegalArgumentException("Null key");
      Collection keyValues = key.getScopesCollection();

      // The passed key doesn't have a parent
      if (keyValues.size() < 2)
         return false;
      
      // If it is a child, it will have one more scope
      if (scopes.size() != keyValues.size() - 1)
         return false;

      Iterator<Scope> thisScopes = scopes.values().iterator();
      Iterator<Scope> keyScopes = keyValues.iterator();
      
      while (thisScopes.hasNext())
      {
         Scope thisScope = thisScopes.next();
         Scope keyScope = keyScopes.next();
         if (thisScope.equals(keyScope) == false)
            return false;
      }
      
      return true;
   }
   
   /**
    * Add a scope
    * 
    * @param scope the scope
    * @return the previous value or null if there wasn't one
    * @throws IllegalArgumentException if scope is null.
    * @throws IllegalStateException if frozen
    */
   public Scope addScope(Scope scope)
   {
      if (scope == null)
         throw new IllegalArgumentException("Null scope");
      if (frozen)
         throw new IllegalStateException("The scope key is frozen");
      if (scopes == EMPTY_SCOPES)
      {
         synchronized (this)
         {
            if (scopes == EMPTY_SCOPES)
               scopes = new ConcurrentSkipListMap<ScopeLevel, Scope>();
         }
      }

      ScopeLevel level = scope.getScopeLevel();
      Scope result = scopes.put(level, scope);
      if (maxScopeLevel == null || level.compareTo(maxScopeLevel) >= 0)
         maxScopeLevel = level;
      return result;
   }
   
   /**
    * Add a scope
    * 
    * @param level the scope level
    * @param qualifier the scope qualifier
    * @return the previous value or null if there wasn't one
    */
   public Scope addScope(ScopeLevel level, Object qualifier)
   {
      Scope scope = new Scope(level, qualifier);
      return addScope(scope);
   }
   
   /**
    * Remove a scope
    * 
    * @param scope the scope
    * @return the previous value or null if there wasn't one
    */
   public Scope removeScope(Scope scope)
   {
      if (scope == null)
         throw new IllegalArgumentException("Null scope");

      return removeScopeLevel(scope.getScopeLevel());
   }
   
   /**
    * Get a scope level
    * 
    * @param scopeLevel the scope level
    * @return the scope or null if there is no such level
    * @throws IllegalArgumentException if parameter key is null    
    */
   public Scope getScopeLevel(ScopeLevel scopeLevel)
   {
      if (scopeLevel == null)
         throw new IllegalArgumentException("Null scope level");
      
      return scopes.get(scopeLevel);
   }
   
   /**
    * Remove a scope level
    * 
    * @param scopeLevel the scope level
    * @return the scope or null if there is no such level
    */
   public Scope removeScopeLevel(ScopeLevel scopeLevel)
   {
      if (scopeLevel == null)
         throw new IllegalArgumentException("Null scope level");
      if (frozen)
         throw new IllegalStateException("The scope key is frozen");

      if (scopes == EMPTY_SCOPES)
         return null;
      
      Scope result = scopes.remove(scopeLevel);
      if (scopeLevel.equals(maxScopeLevel))
      {
         maxScopeLevel = null;
         for (ScopeLevel level : scopes.keySet())
            maxScopeLevel = level; 
      }
      return result;
   }

   public String toString()
   {
      return scopes.values().toString();
   }
   
   public boolean equals(Object object)
   {
      if (object == this)
         return true;
      if (object == null || object instanceof ScopeKey == false)
         return false;
      ScopeKey other = (ScopeKey) object;
      if (other instanceof UnmodifiableScopeKey)
         return other.equals(this);
      else
         return scopes.equals(other.scopes);
   }

   public int hashCode()
   {
      return computeHashCode(scopes.values());  
   }

   /**
    * Clone a ScopeKey instance.
    *
    * @return an unfrozen instance (even if the current instance is frozen).
    */
   public ScopeKey clone()
   {
      try
      {
         ScopeKey result = (ScopeKey) super.clone();
         result.scopes = new ConcurrentSkipListMap<ScopeLevel, Scope>(scopes);
         result.frozen = false;
         return result;
      }
      catch (CloneNotSupportedException e)
      {
         throw new Error(e);
      }
   }

   protected Scope[] getArray()
   {
      Collection<Scope> scopes = getScopes();
      return scopes.toArray(new Scope[0]);
   }

   protected Collection<Scope> getScopesCollection()
   {
      return scopes.values();
   }
   
   protected static int computeHashCode(Scope[] scopeArray)
   {
      return computeHashCode(Arrays.asList(scopeArray));
   }

   protected static int computeHashCode(Iterable<Scope> scopeCollection)
   {
      int hashCode = 0;
      for (Scope scope : scopeCollection)
         hashCode += scope.hashCode();
      return hashCode;
   }
   
   private final static class UnmodifiableConcurrentSkipListMap<K, V> extends ConcurrentSkipListMap<K, V>
   {
      private static final long serialVersionUID = 1L;

      @Override
      public void clear()
      {
         throw new UnsupportedOperationException();
      }

      @Override
      public V put(K key, V value)
      {
         throw new UnsupportedOperationException();
      }

      @Override
      public V putIfAbsent(K key, V value)
      {
         throw new UnsupportedOperationException();
      }

      @Override
      public boolean remove(Object key, Object value)
      {
         throw new UnsupportedOperationException();
      }

      @Override
      public V remove(Object key)
      {
         throw new UnsupportedOperationException();
      }

      @Override
      public boolean replace(K key, V oldValue, V newValue)
      {
         throw new UnsupportedOperationException();
      }

      @Override
      public V replace(K key, V value)
      {
         throw new UnsupportedOperationException();
      }

      @Override
      public void putAll(Map<? extends K, ? extends V> m)
      {
         throw new UnsupportedOperationException();
      }
   }
}
